@@ -445,7 +445,8 @@ static struct button_widget button_widgets[] = {
 };
 #define NBUTTONS (sizeof(button_widgets) / sizeof(struct button_widget))
 
-#define UI_FILE PACKAGE_UI_DIR "/gcalctool.ui"
+#define UI_FILE      PACKAGE_UI_DIR "/gcalctool.ui"
+#define UI_FINC_FILE PACKAGE_UI_DIR "/financial.ui"
 
 #define MAXBITS 64      /* Bit panel: number of bit fields. */
 
@@ -552,6 +553,29 @@ typedef enum {
     POPUP_CENTERED   /* Center popup within baseframe */
 } PopupLocation;
 
+static void load_ui(GtkBuilder *ui, const gchar *filename)
+{
+    GError *error = NULL;
+    GtkWidget *dialog;
+
+    gtk_builder_add_from_file(ui, filename, &error);
+    if (error == NULL)
+        return;
+        
+    dialog = gtk_message_dialog_new(NULL, 0,
+                                    GTK_MESSAGE_ERROR,
+                                    GTK_BUTTONS_NONE,
+                                    /* Translators: Title of the error dialog when unable to load the UI files */
+                                    N_("Error loading user interface"));
+    gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
+                                             /* Translators: Description in UI error dialog when unable to load the UI files. %s is replaced with the error message provided by GTK+ */
+                                             N_("A required file is missing or damaged, please check your installation.\n\n%s"), error->message);
+    gtk_dialog_add_buttons(GTK_DIALOG(dialog), GTK_STOCK_QUIT, GTK_RESPONSE_ACCEPT, NULL);
+    
+    gtk_dialog_run(GTK_DIALOG(dialog));
+    exit(0);
+}
+
 static void set_data(GtkBuilder *ui, const gchar *object_name, const gchar *name, gpointer value)
 {
     g_object_set_data(gtk_builder_get_object(ui, object_name), name, value);
@@ -673,13 +697,13 @@ ui_update_modifier_mode()
 {
     static char *sine_labels[]      = {
         /* Translators: The sine button */
-        N_("Sin"),
+        N_("sin"),
         /* Translators: The inverse sine button */        
-        N_("Sin<sup>-1</sup>"),
+        N_("sin<sup>−1</sup>"),
         /* Translators: The hyperbolic sine button */
-        N_("Sinh"),
+        N_("sinh"),
         /* Translators: The inverse hyperbolic sine button */
-        N_("Sinh<sup>-1</sup>")};
+        N_("sinh<sup>−1</sup>")};
     static char *sine_tooltips[]    = {
         /* Translators: The sine tooltip */
         N_("Sine [k]"),
@@ -692,13 +716,13 @@ ui_update_modifier_mode()
     static int  sine_functions[]    = {FN_SIN, FN_ASIN, FN_SINH, FN_ASINH};
     static char *cosine_labels[]    = {
         /* Translators: The cosine button */
-        N_("Cos"),
+        N_("cos"),
         /* Translators: The inverse cosine button */        
-        N_("Cos<sup>-1</sup>"),
+        N_("cos<sup>−1</sup>"),
         /* Translators: The hyperbolic cosine button */
-        N_("Cosh"),
+        N_("cosh"),
         /* Translators: The inverse hyperbolic cosine button */
-        N_("Cosh<sup>-1</sup>")};
+        N_("cosh<sup>−1</sup>")};
     static char *cosine_tooltips[]  = {
         /* Translators: The cosine tooltip */
         N_("Cosine [j]"),
@@ -711,13 +735,13 @@ ui_update_modifier_mode()
     static int  cosine_functions[]  = {FN_COS, FN_ACOS, FN_COSH, FN_ACOSH};
     static char *tangent_labels[]   = {
         /* Translators: The tangent button */
-        N_("Tan"),
+        N_("tan"),
         /* Translators: The inverse tangent button */        
-        N_("Tan<sup>-1</sup>"),
+        N_("tan<sup>−1</sup>"),
         /* Translators: The hyperbolic tangent button */
-        N_("Tanh"),
+        N_("tanh"),
         /* Translators: The inverse hyperbolic tangent button */
-        N_("Tanh<sup>-1</sup>")};
+        N_("tanh<sup>−1</sup>")};
     static char *tangent_tooltips[] = {
         /* Translators: The tangent tooltip */
         N_("Tangent [w]"),
@@ -731,36 +755,36 @@ ui_update_modifier_mode()
     
     static char *ln_labels[]        = {
         /* Translators: The natural logaritm button */
-        N_("Ln"),
+        N_("ln"),
         /* Translators: The e to the power of x button */
         N_("e<sup><i>x</i></sup>")};
     static char *ln_tooltips[]      = {
         /* Translators: Tooltip for the natural log button */
-        N_("Natural log [n]"),
+        N_("Natural logarithm [n]"),
         /* Translators: Tooltip for the e to the power of x button */
         N_("e to the power of the displayed value [N]")};
     static int ln_functions[]       = {FN_NATURAL_LOGARITHM, FN_E_POW_X};
     
     static char *log_labels[]       = {
         /* Translators: The 10-based logaritm button */
-        N_("Log"),
+        N_("log"),
         /* Translators: The 10 to the power of x button */
         N_("10<sup><i>x</i></sup>")};
     static char *log_tooltips[]     = {
         /* Translators: Tooltip for the log base 10 button */
-        N_("Base 10 log [g]"),
+        N_("Base 10 logarithm [g]"),
         /* Translators: Tooltip for the 10 to the power of x button */
         N_("10 to the power of displayed value [G]")};
     static int log_functions[]      = {FN_LOGARITHM, FN_10_POW_X};
     
     static char *log2_labels[]      = {
         /* Translators: The 2-based logaritm button */
-        N_("Log<sub>2</sub>"),
+        N_("log<sub>2</sub>"),
         /* Translators: The 2 to the power of x button */
         N_("2<sup><i>x</i></sup>")};
     static char *log2_tooltips[]    = {
         /* Translators: Tooltip for the log base 2 button */
-        N_("Base 2 log [h]"),
+        N_("Base 2 logarithm [h]"),
         /* Translators: Tooltip for the 2 to the power of x button */
         N_("2 to the power of the displayed value [H]")};
     static int log2_functions[]     = {FN_LOGARITHM2, FN_2_POW_X};
@@ -1614,19 +1638,10 @@ finc_response_cb(GtkWidget *widget, gint response_id)
 static void
 setup_finc_dialogs(void)
 {
-    GError *error = NULL;
     int i, j;
     
-    // FIXME: Use same code as main UI
     X.financial = gtk_builder_new();
-    gtk_builder_add_from_file(X.financial, PACKAGE_UI_DIR "/financial.ui", &error);
-    if (error != NULL)
-    {
-        g_object_unref(X.financial);
-        X.financial = NULL;
-        g_warning("Error loading financial UI: %s\n", error->message);
-        return;
-    }
+    load_ui(X.financial, UI_FINC_FILE);
     
     set_int_data(X.financial, "ctrm_dialog", "finc_dialog", FINC_CTRM_DIALOG);
     set_int_data(X.financial, "ddb_dialog", "finc_dialog", FINC_DDB_DIALOG);
@@ -2475,28 +2490,9 @@ create_main_window()
     GtkSizeGroup *size_group;
     GtkAccelGroup *accel_group;
     GtkWidget *treeview;
-    GError *error = NULL;
    
     X.ui = gtk_builder_new();
-    gtk_builder_add_from_file(X.ui, UI_FILE, &error);
-    if (error != NULL) {
-        GtkWidget *dialog;
-        
-        dialog = gtk_message_dialog_new(NULL, 0,
-                                        GTK_MESSAGE_ERROR,
-                                        GTK_BUTTONS_NONE,
-                                        /* Translators: Title of the error dialog when unable to load the UI files */
-                                        N_("Error loading user interface"));
-        // FIXME: Use error->message
-        g_warning("Unable to load file %s: %s\n", UI_FILE, error->message);
-        gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
-                                                 /* Translators: Description in UI error dialog when unable to load the UI files. %s is replaced with the path of the missing file */
-                                                 N_("The user interface file %s is missing or unable to be loaded. Please check your installation."), UI_FILE);
-        gtk_dialog_add_buttons(GTK_DIALOG(dialog), GTK_STOCK_QUIT, GTK_RESPONSE_ACCEPT, NULL);
-        
-        gtk_dialog_run(GTK_DIALOG(dialog));
-        exit(0);
-    }
+    load_ui(X.ui, UI_FILE);
     gtk_builder_connect_signals(X.ui, NULL);
 
     X.clipboard_atom   = gdk_atom_intern("CLIPBOARD", FALSE);
